/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import Foundation
import AppKit

import struct BluetoothControl.Logger

class BlackmagicSlider: NSSlider, BlackmagicSliderCellDelegate {
    typealias ActionCallbackFn = (BlackmagicSlider) -> Void
    typealias ImportantValueCallbackFn = (Double) -> Void
    var m_onTentativeValueChanged: ActionCallbackFn?
    var m_onValueChanged: ActionCallbackFn?
    var m_onImportantValue: ImportantValueCallbackFn?
    var m_timer: Timer!
    var m_importantValues = [Double]()
    var m_currentValue: Double = 0.0

    override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        (self.cell as! BlackmagicSliderCell).delegate = self
    }

    required init?(coder: NSCoder) {
        super.init(coder: coder)
        (self.cell as! BlackmagicSliderCell).delegate = self
    }

    func addImportantValues(_ values: [Double]) {
        for value in values {
            m_importantValues.append(value)
        }
    }

    func setNewValue(_ newValue: Float) {
        m_currentValue = Double(newValue)
        self.floatValue = newValue
    }

    @IBInspectable open var trackWidth: CGFloat = 2 {
        didSet { setNeedsDisplay() }
    }

    func setCallbacks(onTentativeValueChanged: ActionCallbackFn? = nil, onValueChanged: @escaping ActionCallbackFn, onImportantValue: ImportantValueCallbackFn? = nil) {
        m_onTentativeValueChanged = onTentativeValueChanged
        m_onValueChanged = onValueChanged
        m_onImportantValue = onImportantValue

        self.target = self
        self.action = #selector(BlackmagicSlider.onSliderChanged)
    }

    func clearCallbacks() {
        m_onTentativeValueChanged = nil
        m_onValueChanged = nil
        m_onImportantValue = nil
        m_timer?.invalidate()
        m_timer = nil

        self.target = nil
        self.action = nil
    }

    func checkForImportantValue(_ sender: BlackmagicSlider) {
        let lastValue = m_currentValue
        let currentValue = Double(sender.floatValue)
        let increasingValue = currentValue > lastValue

        let importantValues = increasingValue ? m_importantValues.sorted(by: { $0 < $1 }) : m_importantValues.sorted(by: { $0 > $1 })
        for importantValue in importantValues {
            let hitImportantValue = (lastValue < importantValue && importantValue < currentValue) || (currentValue < importantValue && importantValue < lastValue)
            if hitImportantValue {
                m_onImportantValue?(importantValue)
            }
        }
    }

    @IBAction func onSliderChanged(_ sender: BlackmagicSlider) {
        m_onTentativeValueChanged?(self)
        checkForImportantValue(sender)
        m_currentValue = Double(sender.floatValue)
    }

    func updateTracking(_: Timer) {
        m_onValueChanged?(self)
    }

    func onStartTracking() {
        self.m_timer = Timer(timeInterval: 0.1, target: self as BlackmagicSlider, selector: #selector(BlackmagicSlider.updateTracking), userInfo: nil, repeats: true)
        m_timer?.tolerance = 0.05
        RunLoop.main.add(self.m_timer, forMode: RunLoopMode.commonModes)
    }

    func onStopTracking() {
        m_onValueChanged?(self)
        m_timer?.invalidate()
        m_timer = nil
    }
}
