/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import AppKit
import BluetoothControl

class MainViewController: NSViewController, IncomingCameraControlToUIDelegate {

	// IBOutlets
    @IBOutlet weak var whiteBalanceLabel: NSTextField!
    @IBOutlet weak var whiteBalanceSlider: BlackmagicSlider!

    @IBOutlet weak var tintLabel: NSTextField!
    @IBOutlet weak var tintSlider: BlackmagicSlider!

    @IBOutlet weak var offSpeedFrameRateLabel: NSTextField!
    @IBOutlet weak var offSpeedFrameRateSlider: BlackmagicSlider!
    @IBOutlet weak var offSpeedFrameRatePrevButton: NSButton!
    @IBOutlet weak var offSpeedFrameRateNextButton: NSButton!
    @IBOutlet weak var offSpeedFrameRateToggleButton: NSButton!

    @IBOutlet weak var irisLabel: NSTextField!
    @IBOutlet weak var irisSlider: BlackmagicSlider!

    @IBOutlet weak var shutterLabel: NSTextField!
    @IBOutlet weak var shutterSlider: BlackmagicSlider!

    @IBOutlet weak var preset1Button: NSButton!
    @IBOutlet weak var preset2Button: NSButton!
    @IBOutlet weak var preset3Button: NSButton!
    @IBOutlet weak var preset4Button: NSButton!
    @IBOutlet weak var preset5Button: NSButton!
    @IBOutlet weak var autoWhiteBalanceButton: NSButton!

    @IBOutlet weak var iso200RadioButton: NSButton!
    @IBOutlet weak var iso400RadioButton: NSButton!
    @IBOutlet weak var iso800RadioButton: NSButton!
    @IBOutlet weak var iso1600RadioButton: NSButton!

	@IBOutlet weak var ISOBox: NSBox!
	
	// Member variables
	weak var m_outgoingCameraControlDelegate: OutgoingCameraControlFromUIDelegate?
	var m_shutterValueIsAngle: Bool = true
	var presetButtons = [NSButton]()
	var isoRadioButtons = [NSButton]()
	
	//==================================================
	//	UIViewController methods
	//==================================================
	override func viewDidLoad() {
        super.viewDidLoad()

        presetButtons.append(preset1Button)
        presetButtons.append(preset2Button)
        presetButtons.append(preset3Button)
        presetButtons.append(preset4Button)
        presetButtons.append(preset5Button)
        presetButtons.append(autoWhiteBalanceButton)

        isoRadioButtons.append(iso200RadioButton)
        isoRadioButtons.append(iso400RadioButton)
        isoRadioButtons.append(iso800RadioButton)
        isoRadioButtons.append(iso1600RadioButton)

		// Set values on sliders.
        whiteBalanceSlider.minValue = Double(VideoConfig.kWhiteBalanceMin)
        whiteBalanceSlider.maxValue = Double(VideoConfig.kWhiteBalanceMax)
        tintSlider.minValue = Double(VideoConfig.kTintMin)
        tintSlider.maxValue = Double(VideoConfig.kTintMax)
        irisSlider.minValue = 0.0
        irisSlider.maxValue = Double(LensConfig.fStopValues.count - 1)
        shutterSlider.minValue = 5.0
        shutterSlider.maxValue = 360.0
        offSpeedFrameRateSlider.minValue = Double(VideoConfig.kOffSpeedFrameRateMin)
        offSpeedFrameRateSlider.maxValue = Double(VideoConfig.kOffSpeedFrameRateMax)
        offSpeedFrameRateSlider.addImportantValues([50.0, 40.0, 30.0])

		// Assign callbacks to our sliders
        whiteBalanceSlider.setCallbacks(onTentativeValueChanged: onWhiteBalanceSliderChanged, onValueChanged: onWhiteBalanceSliderSet)
        tintSlider.setCallbacks(onTentativeValueChanged: onTintSliderChanged, onValueChanged: onTintSliderSet)
        irisSlider.setCallbacks(onTentativeValueChanged: nil, onValueChanged: onIrisSliderSet)
        shutterSlider.setCallbacks(onTentativeValueChanged: onShutterSliderChanged, onValueChanged: onShutterSliderSet)
        offSpeedFrameRateSlider.setCallbacks(onTentativeValueChanged: onOffSpeedFrameRateSliderChanged, onValueChanged: onOffSpeedFrameRateSliderSet, onImportantValue: onOffSpeedFrameRateSliderImportantValue)

		// Register this view controller as a CameraControlToUIDelegate with the CameraControlInterface,
		// and assign CameraControlInterface as our OutgoingCameraControlFromUIDelegate.
        let appDelegate = NSApplication.shared().delegate as! AppDelegate
        let cameraControlInterface: CameraControlInterface = appDelegate.cameraControlInterface
        m_outgoingCameraControlDelegate = cameraControlInterface
        cameraControlInterface.m_cameraControlToUIDelegate = self
        cameraControlInterface.onControlViewLoaded()
    }

	//==================================================
	//	IBActions
	//==================================================
    @IBAction func onPrevWhiteBalanceButtonClicked(_: NSButton) {
        if let newValues: (whiteBalance: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onWhiteBalanceDecremented() {
            updateWhiteBalanceWidgets(newValues.whiteBalance)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onNextWhiteBalanceButtonClicked(_: NSButton) {
        if let newValues: (whiteBalance: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onWhiteBalanceIncremented() {
            updateWhiteBalanceWidgets(newValues.whiteBalance)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onPrevTintButtonClicked(_: NSButton) {
        if let newValues: (tint: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onTintDecremented() {
            updateTintWidgets(newValues.tint)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onNextTintButtonClicked(_: NSButton) {
        if let newValues: (tint: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onTintIncremented() {
            updateTintWidgets(newValues.tint)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onPrevOffSpeedFrameRateButtonClicked(_: NSButton) {
        if let newOffSpeedFrameRate = m_outgoingCameraControlDelegate?.onOffSpeedFrameRateDecremented() {
            updateOffSpeedFrameRateWidgets(newOffSpeedFrameRate, true)
        }
    }

    @IBAction func onNextOffSpeedFrameRateButtonClicked(_: NSButton) {
        if let newOffSpeedFrameRate = m_outgoingCameraControlDelegate?.onOffSpeedFrameRateIncremented() {
            updateOffSpeedFrameRateWidgets(newOffSpeedFrameRate, true)
        }
    }

    @IBAction func offSpeedButtonTapped(_ sender: NSButton) {
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateToggled(sender.isSelected)
    }

    @IBAction func onPrevIrisButtonClicked(_: NSButton) {
        let newIrisIndex = m_outgoingCameraControlDelegate?.onIrisDecremented()
        if newIrisIndex != nil {
            updateIrisWidgets(newIrisIndex!)
        }
    }

    @IBAction func onNextIrisButtonClicked(_: NSButton) {
        let newIrisIndex = m_outgoingCameraControlDelegate?.onIrisIncremented()
        if newIrisIndex != nil {
            updateIrisWidgets(newIrisIndex!)
        }
    }

    @IBAction func onPrevShutterButtonClicked(_: NSButton) {
        let newShutterValue = m_outgoingCameraControlDelegate?.onShutterDecremented()
        if newShutterValue != nil {
            updateShutterWidgets(Float(newShutterValue!))
        }
    }

    @IBAction func onNextShutterButtonClicked(_: NSButton) {
        let newShutterValue = m_outgoingCameraControlDelegate?.onShutterIncremented()
        if newShutterValue != nil {
            updateShutterWidgets(Float(newShutterValue!))
        }
    }

    @IBAction func onISORadioButtonClicked(_ sender: NSButton) {
        let isoIndex: Int = sender.tag
        m_outgoingCameraControlDelegate?.onISOChanged(isoIndex)
        updateISOWidgets(isoIndex)
    }

    @IBAction func onPresetButtonClicked(_ sender: NSButton) {
        let currentlySelected = !sender.isSelected
        let selectedButtons = presetButtons.filter { return $0.isSelected }
        let saveCustomValues = selectedButtons.count != 0
        presetButtons.forEach { $0.isSelected = false }

        let presetIndex = sender.tag
        let newValues: (whiteBalance: Int16, tint: Int16)? = m_outgoingCameraControlDelegate?.onWhiteBalancePresetPressed(presetIndex, currentlySelected, saveCustomValues)
        if newValues != nil {
            updateWhiteBalanceWidgets(newValues!.whiteBalance)
            updateTintWidgets(newValues!.tint)
        }

        sender.isSelected = !currentlySelected
        presetButtons.last!.isSelected = false
    }

	//==================================================
	//	BlackmagicSlider callbacks
	//==================================================
    func onWhiteBalanceSliderChanged(_: BlackmagicSlider) {
        let newWhiteBalance = (Int16(whiteBalanceSlider.floatValue) / VideoConfig.kWhiteBalanceStep) * VideoConfig.kWhiteBalanceStep
        updateWhiteBalanceWidgets(newWhiteBalance)
    }

    func onWhiteBalanceSliderSet(_: BlackmagicSlider) {
        let newWhiteBalance = (Int16(whiteBalanceSlider.floatValue) / VideoConfig.kWhiteBalanceStep) * VideoConfig.kWhiteBalanceStep
        let presetIndex: Int? = m_outgoingCameraControlDelegate?.onWhiteBalanceChanged(newWhiteBalance)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onTintSliderChanged(_: BlackmagicSlider) {
        let newTint = Int16(tintSlider.floatValue)
        updateTintWidgets(newTint)
    }

    func onTintSliderSet(_: BlackmagicSlider) {
        let newTint = Int16(tintSlider.floatValue)
        let presetIndex: Int? = m_outgoingCameraControlDelegate?.onTintChanged(newTint)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onIrisSliderSet(_: BlackmagicSlider) {
        let newIndex = Int(irisSlider.floatValue)
        m_outgoingCameraControlDelegate?.onIrisChanged(newIndex)
        updateIrisWidgets(newIndex)
    }

    func onShutterSliderChanged(_: BlackmagicSlider) {
        let newShutterValue = shutterSlider.floatValue.rounded()
        updateShutterWidgets(newShutterValue)
    }

    func onShutterSliderSet(_: BlackmagicSlider) {
        let newShutterValue = shutterSlider.floatValue.rounded()
        updateShutterWidgets(newShutterValue)
        m_outgoingCameraControlDelegate?.onShutterChanged(Double(newShutterValue))
    }

    func onOffSpeedFrameRateSliderChanged(_: BlackmagicSlider) {
        let newOffSpeedFrameRate = Int16(offSpeedFrameRateSlider.floatValue)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }

    func onOffSpeedFrameRateSliderSet(_: BlackmagicSlider) {
        let newOffSpeedFrameRate = Int16(offSpeedFrameRateSlider.floatValue)
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateChanged(newOffSpeedFrameRate)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }

    func onOffSpeedFrameRateSliderImportantValue(_ value: Double) {
        let newOffSpeedFrameRate = Int16(value)
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateChanged(newOffSpeedFrameRate)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }

	//==================================================
	//	IncomingCameraControlToUIDelegate methods
	//==================================================
    func onWhiteBalanceReceived(_ whiteBalance: Int16, _ tint: Int16, _ presetIndex: Int) {
        updateWhiteBalanceWidgets(whiteBalance)
        updateTintWidgets(tint)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onOffSpeedFrameRateReceived(_ offSpeedEnabled: Bool, _ frameRate: Int16) {
        updateOffSpeedFrameRateWidgets(frameRate, offSpeedEnabled)
    }

//    func onIrisReceived(_ fStopIndex: Int) {
//        updateIrisWidgets(fStopIndex)
//    }

//    func onShutterReceived(_ shutterAngle: Double) {
//        updateShutterWidgets(Float(shutterAngle))
//    }
	
	func onShutterAngleReceived(_ shutterAngle: Double) {
		m_shutterValueIsAngle = true
		shutterSlider.minValue = VideoConfig.kShutterAngleMin
		shutterSlider.maxValue = VideoConfig.kShutterAngleMax
		updateShutterWidgets(Float(shutterAngle))
	}
	
	func onShutterSpeedReceived(_ shutterSpeed: Int32) {
		m_shutterValueIsAngle = false
		shutterSlider.minValue = Double(VideoConfig.kShutterSpeedMin)
		shutterSlider.maxValue = Double(VideoConfig.kShutterSpeedMax)
		updateShutterWidgets(Float(shutterSpeed))
	}

    func onISOReceived(_ isoIndex: Int) {
		applyISOText()
        updateISOWidgets(isoIndex)
    }
	
	func onGainReceived(_ gainIndex: Int) {
		applyGainText()
		updateISOWidgets(gainIndex)
	}
	
	func setIrisControlEnabled(_ enabled: Bool) {}
	func setShutterControlEnabled(_ enabled: Bool) {}
	func onTransportModeReceived(_ transportMode: CCUPacketTypes.MediaTransportMode) {}
	
	//==================================================
	//	UI control
	//==================================================
	func updateWhiteBalanceWidgets(_ whiteBalance: Int16) {
		whiteBalanceLabel.stringValue = "\(whiteBalance)K"
		whiteBalanceSlider.floatValue = Float(whiteBalance)
	}
	
	func updateTintWidgets(_ tint: Int16) {
		tintLabel.stringValue = "\(tint)"
		tintSlider.floatValue = Float(tint)
	}
	
	func updateOffSpeedFrameRateWidgets(_ frameRate: Int16, _ offSpeedEnabled: Bool) {
		offSpeedFrameRateLabel.stringValue = offSpeedEnabled ? "\(frameRate) \(String.Localized("FrameRate.FPS"))" : String.Localized("FrameRate.Sync")
		if offSpeedEnabled {
			offSpeedFrameRateSlider.floatValue = Float(frameRate)
		}
		
		offSpeedFrameRateSlider.isEnabled = offSpeedEnabled
		offSpeedFrameRateNextButton.isEnabled = offSpeedEnabled
		offSpeedFrameRatePrevButton.isEnabled = offSpeedEnabled
		offSpeedFrameRateToggleButton.isSelected = offSpeedEnabled
	}
	
	func updateIrisWidgets(_ newIrisIndex: Int) {
		if newIrisIndex >= 0 && newIrisIndex < LensConfig.fStopValues.count {
			let fStop = LensConfig.fStopValues[newIrisIndex]
			irisLabel.stringValue = LensConfig.GetFStopString(fStop)
			irisSlider.floatValue = Float(newIrisIndex)
		}
	}
	
	func updateShutterWidgets(_ newShutterValue: Float) {
		shutterSlider.floatValue = newShutterValue
		
		if (m_shutterValueIsAngle) {
			shutterLabel.stringValue = "\(newShutterValue.getStringValue())°"
		}
		else {
			shutterLabel.stringValue = "1/\(Int(newShutterValue))"
		}
	}
	
	func updateISOWidgets(_ index: Int) {
		isoRadioButtons.forEach { $0.isSelected = false }
		isoRadioButtons[index].isSelected = true
	}
	
	func applyISOText() {
		ISOBox.title = String.Localized("ISOGain.ISOTitle")
		iso200RadioButton.title = String.Localized("ISOGain.ISO200")
		iso400RadioButton.title = String.Localized("ISOGain.ISO400")
		iso800RadioButton.title = String.Localized("ISOGain.ISO800")
		iso1600RadioButton.title = String.Localized("ISOGain.ISO1600")
	}
	
	func applyGainText() {
		ISOBox.title = String.Localized("ISOGain.GainTitle")
		iso200RadioButton.title = String.Localized("ISOGain.GainNeg6")
		iso400RadioButton.title = String.Localized("ISOGain.Gain0")
		iso800RadioButton.title = String.Localized("ISOGain.Gain6")
		iso1600RadioButton.title = String.Localized("ISOGain.Gain12")
	}
}
