/* -LICENSE-START-
** Copyright (c) 2018 Blackmagic Design
**
** Permission is hereby granted, free of charge, to any person or organization
** obtaining a copy of the software and accompanying documentation covered by
** this license (the "Software") to use, reproduce, display, distribute,
** execute, and transmit the Software, and to prepare derivative works of the
** Software, and to permit third-parties to whom the Software is furnished to
** do so, all subject to the following:
**
** The copyright notices in the Software and this entire statement, including
** the above license grant, this restriction and the following disclaimer,
** must be included in all copies of the Software, in whole or in part, and
** all derivative works of the Software, unless such copies or derivative
** works are solely in the form of machine-executable object code generated by
** a source language processor.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
** DEALINGS IN THE SOFTWARE.
** -LICENSE-END-
*/

import Cocoa
import CoreBluetooth

// Note: For a complete list of Service and Characteristic UUIDs supported by Blackmagic Design cameras,
// please see the README distrubuted with this sample code.

// Service CBUUIDs (see note above)
let kBlackmagicCameraService = CBUUID(string: "291D567A-6D75-11E6-8B77-86F30CA893D3")
let kAdvertisedServiceUUIDs: [CBUUID] = [kBlackmagicCameraService]

// Characteristic CBUUIDs (see note above)
let kOutgoingCCU = CBUUID(string: "5DD3465F-1AEE-4299-8493-D2ECA2F8E1BB")
let kIncomingCCU = CBUUID(string: "B864E140-76A0-416A-BF30-5876504537D9")
let kCameraStatus = CBUUID(string: "7FE8691D-95DC-4FC5-8ABD-CA74339B51B9")
let kMainCharacteristicUUIDs: [CBUUID] = [
	kOutgoingCCU,
	kIncomingCCU,
	kCameraStatus
]

let kCameraPowerOnFlag: UInt8 = 1
let kPowerOn = Data(bytes: [kCameraPowerOnFlag])
let kCCUCommandHeaderSize = 4
let kCUUPayloadOffset = 8

//==================================================
//	ViewController implementation
//==================================================
class ViewController: NSViewController, CBCentralManagerDelegate, CBPeripheralDelegate {
	
	var peripheral: CBPeripheral?			// A local reference to the Blackmagic camera
	var centralManager: CBCentralManager?	// The centralManager searches/discovers/connects to the peripheral (the Blackmagic camera)
	var isRecording: Bool = false
	
	// IBOutlet reference to record button
	@IBOutlet weak var recordButton: NSButton!
	
	//==================================================
	//	NSViewController methods
	//==================================================
	override func viewDidLoad() {
		super.viewDidLoad()
		
		// Create a CBCentralManager, and keep a reference to it.
		// Set this ViewController as its CBCentralManagerDelegate.
		centralManager = CBCentralManager(delegate: self as CBCentralManagerDelegate, queue: nil)
		
		recordButton.isEnabled = false
	}
	
	// IBAction func invoked when record button is clicked
	@IBAction func onRecordClicked(_ sender: Any) {
		if peripheral != nil {
			let shouldRecord = !isRecording
			sendCCUPacket(shouldRecord)
		}
	}
	
	//==================================================
	//	CBCentralManagerDelegate methods
	//==================================================
	func centralManagerDidUpdateState(_ central: CBCentralManager) {
		// Comparing rawValues to work around an SDK change between macOS 10.12 and 10.13.
		// 'central.state' changed types, but the raw values stayed the same. This workaround
		// allows us to support a range of SDKs.
		var powerOn = (central.state.rawValue == CBCentralManagerState.poweredOn.rawValue)

		// If we receive a notification that the CBCentralManager is 'poweredOn', we can start scanning for peripherals.
		// Provide an array containing the Blackmagic Design camera service CBUUID, to ignore non-Blackmagic devices.
		if powerOn {
			centralManager?.scanForPeripherals(withServices: kAdvertisedServiceUUIDs, options: [CBCentralManagerScanOptionAllowDuplicatesKey: true])
		}
	}
	
	func centralManager(_ central: CBCentralManager, didDiscover peripheral: CBPeripheral, advertisementData: [String : Any], rssi RSSI: NSNumber) {
		// We have discovered a Blackmagic Design camera, we can now connect to it.
		// This example connects to the first camera it discovers. Uncomment the line below to connect to a specific camera.
		// You'll need to change the string from "A:2E8945C9" to the name found in your camera's Bluetooth Setup menu.
		
		if self.peripheral == nil {
			//let shouldConnect = true
			let shouldConnect = (peripheral.name != nil) ? peripheral.name!.contains("A:2E8945C9") : false
			
			if shouldConnect {
				self.peripheral = peripheral
				centralManager?.connect(peripheral, options: nil)
			}
		}
	}
	
	func centralManager(_ central: CBCentralManager, didConnect peripheral: CBPeripheral) {
		// We connected successfully to a camera. Start discovering services.
		self.peripheral = peripheral
		self.peripheral?.delegate = self
		self.peripheral?.discoverServices(kAdvertisedServiceUUIDs)
	}
	
	func centralManager(_: CBCentralManager, didFailToConnect peripheral: CBPeripheral, error: Error?) {
		print("Failed to connect to \(String(describing: peripheral.name)), error: \(error!)")
	}
	
	//==================================================
	//	CBPeripheralDelegate methods
	//==================================================
	func peripheral(_ peripheral: CBPeripheral, didDiscoverServices error: Error?) {
		// We discovered a service. Start discovering characteristics.
		if error != nil {
			print("didDiscoverServices, error: \(error!)")
			return
		}
		
		if let services = peripheral.services {
			for service in services {
				if service.uuid == kBlackmagicCameraService {
					peripheral.discoverCharacteristics(kMainCharacteristicUUIDs, for: service)
				}
			}
		}
	}
	
	func peripheral(_ peripheral: CBPeripheral, didDiscoverCharacteristicsFor service: CBService, error: Error?) {
		// We discovered characteristics for a particular service.
		
		if error != nil {
			print("didDiscoverCharacteristicsFor, error: \(error!)")
			return
		}
		
		if let characteristics = service.characteristics {
			for characteristic in characteristics {
				switch(characteristic.uuid) {
				case kCameraStatus:
					// When we discover the Camera Status characteristic, write to it (in this case we are turning it on).
					// As this characteristic is encrypted, this command will initiate a bonding request between this cental
					// and the peripheral. A six-digit pin will appear in your camera's Bluetooth Setup menu, and macOS
					// will prompt you to enter the pin.
					peripheral.writeValue(kPowerOn, for: characteristic, type: CBCharacteristicWriteType.withResponse)
				default:
					break
				}
				
				// Request notifications from characteristics, so we receive callbacks when values are changed.
				// After this call, we will start receiving CCU commands and other Bluetooth updates from the camera
				// (when we are successfully bonded).
				peripheral.setNotifyValue(true, for: characteristic)
			}
		}
	}
	
	func peripheral(_ peripheral: CBPeripheral, didUpdateValueFor characteristic: CBCharacteristic, error: Error?) {
		// We receive all Bluetooth notfications from the camera through this delegate method.
	
		if error != nil {
			print("didDiscoverCharacteristicsFor, error: \(error!)")
			return
		}
		
		switch(characteristic.uuid) {
		case kIncomingCCU:
			onCCUPacketReceived(data: characteristic.value)
			break
		default:
			break
		}
	}
	
	//==================================================
	//	Custom CCU methods
	//==================================================
	func onCCUPacketReceived(data: Data?) {
		// We received a CCU packet from the camera
		if let data = data {
			
			// Convert the data into a byte array
			let byteArray = [UInt8](data)
			
			// Make sure it is of a sane size.
			let packetSize = UInt8(byteArray.count)
			let isSizeValid = (packetSize >= 8 && packetSize <= 64)
			if !isSizeValid {
				return
			}
			
			// Read the values from the byte array. Consult the Blackmagic Design CCU Protocol for details
			// of specific CCU commands.
			let category = byteArray[4]
			let parameter = byteArray[5]
			
			// In this example, we are only interested in the CCU command used to record and stop recording.
			let kTransportCategory: UInt8 = 10
			let kTransportModeParameter: UInt8 = 1
			if category == kTransportCategory && parameter == kTransportModeParameter {
				let transportMode = byteArray[kCUUPayloadOffset]
				isRecording = (transportMode == 2)
				recordButton.title = isRecording ? "Stop" : "Record"
			}
			
			recordButton.isEnabled = true
		}
	}
	
	func sendCCUPacket(_ shouldRecord: Bool) {
		
		// Create our payload. This payload is implementing the Transport Mode CCU command.
		// We'll ignore all but the first element in this example.
		let payloadSize = 5
		var payload = [UInt8](repeating: 0, count: payloadSize)
		payload[0] = shouldRecord ? 2 : 0	// transport mode (0 = preview, 1 = playback, 2 = record)
		payload[1] = UInt8(bitPattern: 1)	// playback speed
		payload[2] = 0						// flags
		payload[3] = 0						// slot 1 storage medium
		payload[4] = 0						// slot 2 storage medium
		
		// Create our CCU command buffer.
		
		// The headers are 8 bytes, and our Transport Mode payload is 5 bytes - 13 bytes in total.
		// CCU commands must be aligned to a 4-byte boundary, so we need to pad this command to a total of 16 bytes.
		let headersSize = 8
		let padBytes = ((payloadSize + 3) & ~3) - payloadSize // this calculates the padding
		let totalCommandLength = Int(headersSize + payloadSize + padBytes)
		var command: [UInt8] = [UInt8](repeating: 0, count: totalCommandLength)
		let length: UInt8 = UInt8(kCCUCommandHeaderSize + payloadSize) // totalCommandLength without the inital 4-byte header
		
		command[0] = 255 	// Broadcast target
		command[1] = length // length
		command[2] = 0 		// command id (0 = change configuration)
		command[3] = 0 		// reserved
		command[4] = 10 	// category
		command[5] = 1 		// parameter
		command[6] = 1 		// data type (1 = uint8)
		command[7] = 0 		// operation type (0 = assign value)
		
		// Add the payload bytes to our command buffer.
		for i in 0 ..< payloadSize {
			command[8 + i] = payload[i]
		}
		
		// Get the characteristic for sending outgoing CCU messages.
		if let characteristic = getCharacteristicForService(kBlackmagicCameraService, kOutgoingCCU) {
			// Convert the command buffer into a Data instance, and send it with CBPeripheral.writeValue()
			let data = Data(bytes: command)
			peripheral?.writeValue(data, for: characteristic, type: CBCharacteristicWriteType.withResponse)
		}
	}
	
	// Returns the CBCharacteristic with the provided characteristic CBUUID, from the CBService with the provided
	// service CBUUID.
	func getCharacteristicForService(_ service: CBUUID, _ characteristic: CBUUID) -> (CBCharacteristic?) {
		var validService: CBService?
		
		if let services = peripheral?.services {
			let filteredServices: [CBService] = services.filter { $0.uuid == service }
			validService = filteredServices.count > 0 ? filteredServices[0] : nil
		}
		
		if let validService = validService {
			if let characteristics = validService.characteristics {
				let filteredCharacteristics: [CBCharacteristic] = characteristics.filter { $0.uuid == characteristic }
				return filteredCharacteristics.count > 0 ? filteredCharacteristics[0] : nil
			}
		}
		
		return nil
	}
}

