/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import Foundation

struct CCUDecodingFunctions {
    static func DecodeCCUPacket(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let byteArray = [UInt8](data)
        let isValid: Bool = CCUValidationFunctions.ValidateCCUPacket(packetAsByteArray: byteArray)

        if isValid {
            let commandLength = byteArray[CCUPacketTypes.PacketFormatIndex.CommandLength]
            let category = CCUPacketTypes.Category(rawValue: byteArray[CCUPacketTypes.PacketFormatIndex.Category])!
            let parameter = byteArray[CCUPacketTypes.PacketFormatIndex.Parameter]

            let dataLength = commandLength - CCUPacketTypes.kCCUCommandHeaderSize
            let payloadOffset = CCUPacketTypes.kCUUPayloadOffset
            let payloadRange: Range<Data.Index> = Data.Index(payloadOffset) ..< Data.Index(payloadOffset + dataLength)
            let payloadData = data.subdata(in: payloadRange)

            try DecodePayloadData(for: category, parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        } else {
            throw CCUDecodingError.PacketValidationFailed
        }
    }

    static func DecodePayloadData(for category: CCUPacketTypes.Category, parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        switch category
        {
        case .Lens:
            try DecodeLensCategory(parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        case .Video:
            try DecodeVideoCategory(parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        case .Status:
            try DecodeStatusCategory(parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        case .Media:
            try DecodeMediaCategory(parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        case .Metadata:
            try DecodeMetadataCategory(parameter: parameter, payloadData: payloadData, respondTo: packetDecodedDelegate)
        default:
            break
        }
    }

    static func ConvertPayloadData<T>(from data: Data) throws -> [T] {
        let typeSize = MemoryLayout<T>.size
        let byteCount = data.count

        let conversionClosure = {
            (ptr: UnsafePointer<T>) -> [T] in
            let ptrToArrayOfType = UnsafeBufferPointer<T>(start: ptr, count: byteCount / typeSize)
            let arrayOfType = [T](ptrToArrayOfType)
            return arrayOfType
        }
        let payload: [T] = data.withUnsafeBytes(conversionClosure)
        return payload
    }

    static func ConvertPayloadDataWithExpectedCount<T>(from data: Data, expectedCount: Int) throws -> [T] {
        let typeSize = MemoryLayout<T>.size
        let byteCount = data.count
        if typeSize > byteCount {
            Logger.LogWithInfo("Payload type size (\(typeSize)) is smaller than data size (\(byteCount))")
            throw CCUDecodingError.PayloadConversionFailed
        }

        let convertedCount = byteCount / typeSize
        if expectedCount != convertedCount {
            Logger.LogWithInfo("Payload expected count (\(expectedCount)) not equal to converted count (\(convertedCount))")
            throw CCUDecodingError.PayloadConversionFailed
        }

        let payload: [T] = data.withUnsafeBytes { (ptr: UnsafePointer<T>) -> [T] in
            return [T](UnsafeBufferPointer<T>(start: ptr, count: convertedCount))
        }

        return payload
    }

    static func ConvertPayloadDataToString(from data: Data) throws -> String {
        let string = String(data: data, encoding: String.Encoding.utf8)
        return string ?? ""
    }

    // Lens Category decoding functions
    static func DecodeLensCategory(parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let parameterType = CCUPacketTypes.LensParameter(rawValue: parameter)
        if parameterType != nil {
            switch parameterType!
            {
            case .ApertureFstop:
                try DecodeApertureFStop(data: payloadData, respondTo: packetDecodedDelegate)
            default:
                break
            }
        } else {
            Logger.LogWithInfo("Invalid value for LensParameter: \(parameter).")
            throw CCUGeneralError.InvalidParameter
        }
    }

    static func DecodeApertureFStop(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int16] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
        let apertureNumber: Int16 = data[0]
        
        var fStopIndex = -1
        if apertureNumber != CCUPacketTypes.kLensAperture_NoLens {
            fStopIndex = LensConfig.GetIndexForApertureNumber(apertureNumber)
        }
        
        packetDecodedDelegate.onIrisReceived(fStopIndex)
    }

    // Video Category decoding functions
    static func DecodeVideoCategory(parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let parameterType = CCUPacketTypes.VideoParameter(rawValue: parameter)
        if parameterType != nil {
            switch parameterType!
            {
            case .SensorGain:
                try DecodeSensorGain(data: payloadData, respondTo: packetDecodedDelegate)
            case .ManualWB:
                try DecodeManualWB(data: payloadData, respondTo: packetDecodedDelegate)
            case .Exposure:
                try DecodeExposure(data: payloadData, respondTo: packetDecodedDelegate)
            case .RecordingFormat:
				try DecodeRecordingFormat(data: payloadData, respondTo: packetDecodedDelegate)
			case .AutoExposureMode:
				try DecodeAutoExposureMode(data: payloadData, respondTo: packetDecodedDelegate)
			case .ShutterAngle:
				try DecodeShutterAngle(data: payloadData, respondTo: packetDecodedDelegate)
			case .ShutterSpeed:
				try DecodeShutterSpeed(data: payloadData, respondTo: packetDecodedDelegate)
			case .Gain:
				try DecodeGain(data: payloadData, respondTo: packetDecodedDelegate)
			case .ISO:
				try DecodeISO(data: payloadData, respondTo: packetDecodedDelegate)
            default:
                break
            }
        } else {
            Logger.LogWithInfo("Invalid value for VideoParameter: \(parameter).")
            throw CCUGeneralError.InvalidParameter
        }
    }

    static func DecodeSensorGain(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
        let sensorGainValue = Int(data[0])
        let sensorGain = sensorGainValue * Int(VideoConfig.kReceivedSensorGainBase)
        packetDecodedDelegate.onISOReceived(sensorGain)
    }

    static func DecodeManualWB(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int16] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 2)

        let whiteBalance: Int16 = data[0]
        let tint: Int16 = data[1]

        packetDecodedDelegate.onWhiteBalanceReceived(whiteBalance, tint)
    }

    static func DecodeExposure(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int32] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
        let shutterSpeed = data[0]
        packetDecodedDelegate.onExposureReceived(shutterSpeed)
    }

    static func DecodeRecordingFormat(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int16] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 5)

        var recordingFormatData = CCUPacketTypes.RecordingFormatData()
        recordingFormatData.frameRate = data[0]
        recordingFormatData.offSpeedFrameRate = data[1]
        recordingFormatData.width = data[2]
        recordingFormatData.height = data[3]

        let flags = UInt16(data[4])
        recordingFormatData.mRateEnabled = (flags & CCUPacketTypes.VideoRecordingFormat.FileMRate.rawValue) > 0
        recordingFormatData.offSpeedEnabled = (flags & CCUPacketTypes.VideoRecordingFormat.SensorOffSpeed.rawValue) > 0
        recordingFormatData.interlacedEnabled = (flags & CCUPacketTypes.VideoRecordingFormat.Interlaced.rawValue) > 0
        recordingFormatData.windowedModeEnabled = (flags & CCUPacketTypes.VideoRecordingFormat.WindowedMode.rawValue) > 0

        packetDecodedDelegate.onRecordingFormatReceived(recordingFormatData)
    }
	
	static func DecodeAutoExposureMode(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
		let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
		if let autoExposureMode = CCUPacketTypes.AutoExposureMode(rawValue: data[0]) {
			packetDecodedDelegate.onAutoExposureModeReceived(autoExposureMode)
		}
	}
	
	static func DecodeShutterAngle(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
		let data: [Int32] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
		let shutterAngleX100 = data[0]
		packetDecodedDelegate.onShutterAngleReceived(shutterAngleX100)
	}
	
	static func DecodeShutterSpeed(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
		let data: [Int32] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
		let shutterSpeed = data[0]
		packetDecodedDelegate.onShutterSpeedReceived(shutterSpeed)
	}
	
	static func DecodeGain(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
		let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
		let gain = data[0]
		packetDecodedDelegate.onGainReceived(Int(gain))
	}
	
	static func DecodeISO(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
		let data: [Int32] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 1)
		let iso = data[0]
		packetDecodedDelegate.onISOReceived(Int(iso))
	}

    // Status Category decoding functions
    static func DecodeStatusCategory(parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let parameterType = CCUPacketTypes.StatusParameter(rawValue: parameter)
        if parameterType != nil {
            switch parameterType!
            {
            case .MediaStatus:
                try DecodeMediaStatus(data: payloadData, respondTo: packetDecodedDelegate)
            case .RemainingRecordTime:
                try DecodeRemainingRecordTime(data: payloadData, respondTo: packetDecodedDelegate)
            default:
                break
            }
        } else {
            Logger.LogWithInfo("Invalid value for StatusParameter: \(parameter).")
            throw CCUGeneralError.InvalidParameter
        }
    }

    static func DecodeMediaStatus(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadData(from: data)
        let slotCount = min(data.count, 2)

        var mediaStatuses: [CCUPacketTypes.MediaStatus] = Array(repeating: CCUPacketTypes.MediaStatus.None, count: 2)
        for slotIndex in 0 ..< slotCount {
            let slotStatus = CCUPacketTypes.MediaStatus(rawValue: data[slotIndex])
            if slotStatus != nil {
                mediaStatuses[slotIndex] = slotStatus!
            }
        }

        packetDecodedDelegate.onMediaStatusReceived(mediaStatuses)
    }

    static func DecodeRemainingRecordTime(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int16] = try ConvertPayloadData(from: data)

        let noCardString = String.Localized("Transport.NoCard")
        var remainingTimesStrings: [String] = [noCardString, noCardString]
        var remaingTimesInMinutes: [Int16] = [0, 0]
        let slotCount = min(data.count, 2)
        for slotIndex in 0 ..< slotCount {
            let remainingTimeValue = data[slotIndex]
            if remainingTimeValue > 0 {
                if remainingTimeValue >= 60 {
                    let min = String.Localized("Transport.Min")
                    remainingTimesStrings[slotIndex] = "\(remainingTimeValue / 60)\(min)"
                    remaingTimesInMinutes[slotIndex] = remainingTimeValue / 60
                } else {
                    let sec = String.Localized("Transport.Sec")
                    remainingTimesStrings[slotIndex] = "\(remainingTimeValue)\(sec)"
                    remaingTimesInMinutes[slotIndex] = remainingTimeValue / 60
                }
            } else if remainingTimeValue < 0 {
                let min = String.Localized("Transport.Min")
                remainingTimesStrings[slotIndex] = "\(-remainingTimeValue)\(min)"
                remaingTimesInMinutes[slotIndex] = -remainingTimeValue
            } else {
                remainingTimesStrings[slotIndex] = String.Localized("Transport.Full")
            }
        }
        
        packetDecodedDelegate.onRecordTimeRemainingReceived(remainingTimesStrings, remaingTimesInMinutes)
    }

    // Media Category decoding functions
    static func DecodeMediaCategory(parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let parameterType = CCUPacketTypes.MediaParameter(rawValue: parameter)

        if parameterType != nil {
            switch parameterType!
            {
            case .TransportMode:
                try DecodeTransportMode(data: payloadData, respondTo: packetDecodedDelegate)
            default:
                break
            }
        } else {
            Logger.LogWithInfo("Invalid value for MediaParameter: \(parameter).")
            throw CCUGeneralError.InvalidParameter
        }
    }

    static func DecodeTransportMode(data: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: data, expectedCount: 5)

        let mode = CCUPacketTypes.MediaTransportMode(rawValue: UInt8(bitPattern: data[0]))
        let storageMediumDisk1 = CCUPacketTypes.ActiveStorageMedium(rawValue: UInt8(bitPattern: data[3]))
		let storageMediumDisk2 = CCUPacketTypes.ActiveStorageMedium(rawValue: UInt8(bitPattern: data[4]))

        if let mode = mode, let storageMediumDisk1 = storageMediumDisk1, let storageMediumDisk2 = storageMediumDisk2 {
            let speed = data[1]
            let flags = UInt8(bitPattern: data[2])

            let loop = (flags & CCUPacketTypes.MediaTransportFlag.Loop.rawValue) > 0
            let playAll = (flags & CCUPacketTypes.MediaTransportFlag.PlayAll.rawValue) > 0
            let disk1Active = (flags & CCUPacketTypes.MediaTransportFlag.Disk1Active.rawValue) > 0
            let disk2Active = (flags & CCUPacketTypes.MediaTransportFlag.Disk2Active.rawValue) > 0
            let timelapseRecording = (flags & CCUPacketTypes.MediaTransportFlag.TimelapseRecording.rawValue) > 0

            var transportInfo = TransportInfo()
            transportInfo.transportMode = mode
            transportInfo.speed = speed
            transportInfo.loop = loop
            transportInfo.playAll = playAll
            transportInfo.disk1Active = disk1Active
            transportInfo.disk2Active = disk2Active
            transportInfo.timelapseRecording = timelapseRecording
            transportInfo.storageMediumDisk1 = storageMediumDisk1
			transportInfo.storageMediumDisk2 = storageMediumDisk2

            packetDecodedDelegate.onTransportInfoReceived(transportInfo)
        } else {
            Logger.LogError("Decoding transport mode failed. mode = \(String(describing: mode))")
        }
    }

    // Media Category decoding functions
    static func DecodeMetadataCategory(parameter: UInt8, payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let parameterType = CCUPacketTypes.MetadataParameter(rawValue: parameter)
        if parameterType != nil {
            switch parameterType!
            {
            case .Reel:
                try DecodeReel(payloadData, respondTo: packetDecodedDelegate)
            case .SceneTags:
                try DecodeSceneTags(payloadData, respondTo: packetDecodedDelegate)
            case .Scene:
                try DecodeScene(payloadData, respondTo: packetDecodedDelegate)
            case .Take:
                try DecodeTake(payloadData, respondTo: packetDecodedDelegate)
            case .GoodTake:
                try DecodeGoodTake(payloadData, respondTo: packetDecodedDelegate)
            case .SlateForType:
                try DecodeSlateForType(payloadData, respondTo: packetDecodedDelegate)
            case .SlateForName:
                try DecodeSlateForName(payloadData, respondTo: packetDecodedDelegate)
            default:
                break
            }
        } else {
            Logger.LogWithInfo("Invalid value for MediaParameter: \(parameter).")
            throw CCUGeneralError.InvalidParameter
        }
    }

    // Metadata decoding functions
    // Important: All metadata CCU commands are undocumented and are subject to change in a future release
    
    static func DecodeReel(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int16] = try ConvertPayloadDataWithExpectedCount(from: payloadData, expectedCount: 1)
        let reelNumber: Int16 = data[0]

        packetDecodedDelegate.onReelReceived(reelNumber)
    }

    static func DecodeSceneTags(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: payloadData, expectedCount: 3)
        let sceneTag = CCUPacketTypes.MetadataSceneTag(rawValue: data[0])
        let locationType = CCUPacketTypes.MetadataLocationTypeTag(rawValue: UInt8(bitPattern: data[1]))
        let dayOrNight = CCUPacketTypes.MetadataDayNightTag(rawValue: UInt8(bitPattern: data[2]))

        if sceneTag != nil && locationType != nil && dayOrNight != nil {
            packetDecodedDelegate.onSceneTagsReceived(sceneTag!, locationType!, dayOrNight!)
        } else {
            Logger.LogWithInfo("Failed enum cast: MetadataSceneTag (\(data[0])), MetadataLocationTypeTag (\(data[1])), MetadataDayNightTag (\(data[2])).")
            throw CCUDecodingError.EnumCastFailed
        }
    }

    static func DecodeScene(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let sceneString: String = try ConvertPayloadDataToString(from: payloadData)
        packetDecodedDelegate.onSceneReceived(sceneString)
    }

    static func DecodeTake(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: payloadData, expectedCount: 2)
        let takeNumber: Int8 = data[0]
        let takeTag = CCUPacketTypes.MetadataTakeTag(rawValue: data[1])
        if takeTag != nil {
            packetDecodedDelegate.onTakeReceived(takeNumber, takeTag!)
        } else {
            Logger.LogWithInfo("Failed enum cast: MetadataTakeTag (\(data[1])).")
            throw CCUDecodingError.EnumCastFailed
        }
    }

    static func DecodeGoodTake(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadDataWithExpectedCount(from: payloadData, expectedCount: 1)
        let goodTake: Int8 = data[0]

        packetDecodedDelegate.onGoodTakeReceived(goodTake == 1)
    }

    static func DecodeSlateForType(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let data: [Int8] = try ConvertPayloadData(from: payloadData)
        if data.count < 1 {
            throw CCUDecodingError.PayloadConversionFailed
        }

        let slateForType: CCUPacketTypes.MetadataSlateForType? = CCUPacketTypes.MetadataSlateForType(rawValue: data[0])
        if slateForType == nil {
            throw CCUDecodingError.PayloadConversionFailed
        }

        packetDecodedDelegate.onSlateForTypeReceived(slateForType!)
    }

    static func DecodeSlateForName(_ payloadData: Data, respondTo packetDecodedDelegate: PacketDecodedDelegate) throws {
        let slateForName: String = try ConvertPayloadDataToString(from: payloadData)
        packetDecodedDelegate.onSlateForNameReceived(slateForName)
    }

    // Timecode decoding
    static let kTimecodeDigits = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
    static let kTimecodeSize = MemoryLayout<UInt32>.size

    static func TimecodeToString(timecode: UInt32) -> String {
        var str: String = ""
        let digitCount = kTimecodeSize * 2
        var shift: Int = kTimecodeSize * 8 - 4

        // Convert BCD timecode to string
        for i in 0 ..< digitCount {
            let digit: Int = (Int(timecode) >> shift) & 0xF

            if digit < 10 {
                str += kTimecodeDigits[digit]
            } else {
                str += "-"
            }

            if (i % 2 == 1) && (i < digitCount - 1) {
                str += ":"
            }

            shift -= 4
        }

        return str
    }
}
